use std::pin::Pin;
use std::task::{Context, Poll};

use futures::prelude::*;
use futures_signals::signal::{Mutable, Signal, MutableSignalCloned};

use termion::event::Key;

use crate::input::Inputs;
use crate::schema::Api;

#[derive(PartialEq, Eq, Debug, Clone)]
pub enum Window {
    Main,
    Help,
}

enum ConnState<F> {
    Connecting(F),
    Connected(Api)
}

/// Application state struct
pub struct Sute<S, F> {

    // TODO: BE SMART. Inputs change the state, resize signals change the state, futures completing
    // change the state.

    pub state: Mutable<SuteState>,
    statesig: MutableSignalCloned<SuteState>,
    signal: S,
    inputs: Inputs,
    api: Option<ConnState<F>>,
}

impl<S: Unpin, F: Unpin> Sute<S, F> {
    pub fn new(s: S, api: F) -> Self {
        let inputs = Inputs::new();
        let state = Mutable::new(SuteState::new());

        Self { 
            statesig: state.signal_cloned(),
            state: state,
            signal: s,
            inputs: inputs,
            api: Some(ConnState::Connecting(api)),
        }
    }

    fn handle_resize(&mut self, new_size: (u16,u16)) {
        (self.state.lock_mut()).size = new_size;
    }

    fn handle_input(&mut self, key: Key) {
        // TODO modify signal implementation so we don't have to modify the state on all ticks.
        let mut state = self.state.lock_mut();
        state.tick = state.tick + 1;
        match key {
            Key::Char('q') => state.running = false,
            Key::Char('?') => state.active_win = Window::Help,
            Key::Esc => {
                if state.active_win == Window::Help {
                    state.active_win = Window::Main;
                }
            }
            _ => {}
        }
    }
}

impl<S: Signal<Item=(u16,u16)> + Unpin, F: Future<Output=Api> + Unpin> Signal for Sute<S, F> {
    type Item = SuteState;
    fn poll_change(mut self: Pin<&mut Self>, cx: &mut Context) -> Poll<Option<Self::Item>> {
        if let Poll::Ready(Some(key)) = Pin::new(&mut self.inputs).poll_next(cx) {
            self.handle_input(key);
        }
        if let Poll::Ready(Some(size)) = Pin::new(&mut self.signal).poll_change(cx) {
            self.handle_resize(size);
        }
        if let Some(ConnState::Connecting(mut apif)) = self.api.take() {
            if let Poll::Ready(api) = Pin::new(&mut apif).poll(cx) {
                self.api = Some(ConnState::Connected(api));
            } else {
                self.api = Some(ConnState::Connecting(apif));
            }
        }

        // TODO chunk this?

        Pin::new(&mut self.statesig).poll_change(cx)
    }
}

#[derive(Debug, Clone)]
pub struct SuteState {
    pub active_win: Window,
    pub size: (u16,u16),
    pub running: bool,
    pub tick: usize,
    pub server: Option<String>,
}

impl SuteState {
    pub fn new() -> Self {
        Self {
            active_win: Window::Main,
            size: (80,20),
            running: true,
            tick: 0,
            server: None
        }
    }
}
